/**************************************************************************/
/*                                                                        */
/*   Program Name :  datachk                                              */
/*                                                                        */
/*      Version:  0.00  2001-08-17                                        */
/*                2.10  2002-06-04                                        */
/*                2.11  2002-10-21  to allow extended file                */
/*                2.20  2004-08-12  add sampling statistics               */
/*                2.21  2005-06-12  minor update                          */
/*                2.30  2005-10-05  achive speed up by factor upto 3      */
/*                2.31  2005-10-07  more faster 100th times               */
/*                2.32  2005-10-11  a bit faster in statistics mode       */
/*                2.40  2006-01-02  support K5/VSSP32                     */
/*                2.41  2006-02-01  minor debug                           */
/*                2.42  2006-07-23  minor debug                           */
/*                2.43  2006-10-27  modify chkfilen and chkfilen2 to      */
/*                                  return # of bad frames                */
/*                2.44  2006-11-16  header separation detection in VSSP32 */
/*                2.45  2006-11-18  add output summary to log file        */
/*                2.46  2007-02-07  add EFLG count                        */
/*                2.50  2007-02-12  add error log and error occured data  */
/*                                   keep mode. suppress unecessary info  */
/*                2.51  2007-02-14  name of "header separation" is        */
/*                                  changed to AUX headder error          */
/*                2.52  2007-02-21  minor update                          */
/*                2.53  2007-02-25  minor update                          */
/*                2.54  2007-02-26  debug in fwd1secK5.  Change pick up   */
/*                                  frame for sampling info to 1st        */
/*                                 (so far 2nd frame info was used)       */
/*                2.60  2007-03-02  more robust against in case of only   */
/*                                  1bit reversal in a sync field         */
/*                2.61 2007-04-24  libvssp32.h was divided to *.c and *.h */
/*                2.62 2007-08-23  use "parseADdata_3N_without_check"     */
/*                                 instead of "parseADdata_3N" to avoid   */
/*                                 unnecessary sync detection message     */
/*                2.70 2007-09-25  add new sampling statistics mode that  */
/*                                 counts occurence by level              */
/*                2.71 2007-10-20  bug in fwd1secK5 function in           */
/*                                 libvssp32.c was fixed                  */
/*                2.72 2007-10-23  minor revision                         */
/*                2.73 2008-02-19  options '--help' and '--ver[sion]'     */
/*                                 were added                             */
/*                2.74 2009-07-31  version change to just reflect         */
/*                                 libvssp32.c update                     */
/*                2.75 2009-10-15  debug in "chkfilen"                    */
/*                2.76 2011-02-07  all "long int" were changed to "int"   */
/*                                                                        */
/*      Data Check for PC-VSSP(IP-VLBI) board                             */
/*                                                                        */
/*                                                                        */
/*      Copyright (c) 2002-2011 T.Kondo/NICT All Right Reserved           */
/*                                                                        */
/**************************************************************************/
/*
      usage
             datachk file_name [mode [logfile [errlog [keepmode]]]] [options]
                  where   file_name -- K5/VSSP (IP-VLBI) data file name
                          mode     --  0: Just check # of bits and time (default)
                                          only error occured frame is diplayed.
                                       1: Display sampling statistics
                                       2: same as mode 0 but each frame is displayed
                                       3: Display sampling statistics by AD level (full)
                        logfile   ---  summary out file name for mode 0 (default: no out).
                                        file name start with '-' means append mode open
                         errlog   ---  log file name to contain error 
                                        detected data file name for mode 0.
                                        This file will not be created in case of no error. 
                                        (default: no log)
                                        file name start with '-' means append mode open
                         keepmode   ---  data file keep mode for error occurred file for mode 0.
                                        0: nothing to do (default)
                                        1: rename to original+.NNNN.err
                                        2: copy to original+.NNNN.err
                                            where NNNN is 0001 to 9999 

        ver 2.11 2002-10-21  support data file created under file extension mode
                             sampling
        ver 2.20 2004-08-12  add display mode that do statistics about sampling biases
        Ver 2.21 2005.06.12  update due to update in common header name
        Ver 2.30 2005.10.05  achieve speed up by factor 1.5 to 3
        Ver 2.31 2005.10.07  more faster by factor 100 or more for bit count
        Ver 2.32 2005.10.11  a bit faster in statistics mode
        Ver 2.40 2006.01.02  support K5/VSSP32
        Ver 2.41 2006.02.01  minor debug (host name disappeared in case of filter through)
        Ver 2.42 2006.07.23  minor debug (correct "datachk32" -> "datachk")
        Ver 2.43 2006.10.27  modify chkfilen and chkfilen2 to return # of bad frames
        Ver 2.44 2006.11.16  header separation detection in VSSP32
        Ver 2.45 2006.11.18  add output summary to log file
        Ver 2.46 2007.02.07  add EFLG count
        Ver 2.47 2007.02.08  add year, total day out for VSSP32
        Ver 2.50 2007-02-12  add error log and error occured data keep mode
                             suppress unnessary info at default mode
        Ver 2.51 2007-02-14  name of "header separation" is changed to AUX headder error
        Ver 2.52 2007-02-21  minor update
        Ver 2.53 2007-02-25  minor update
        Ver 2.60 2007-03-02  more robust sync finding was adopted.
                               1) allow 1bit pattern reversal in a sync field
                               2) use sampling parameters of 1st frame as sync pattern
                               3) use time information to recognize header 
        Ver 2.61 2007-04-24  libvssp32.h was divided to *.c and *.h
        Ver 2.62 2007-08-23  use "parseADdata_3N_without_check"     
                             instead of "parseADdata_3N" to avoid   
                             unnecessary sync detection message
        Ver 2.70 2007-09-25  add new sampling statistics mode that  
                                 counts occurence by level              
        Ver 2.71 2007-10-20  bug in fwd1secK5 function in libvssp32.c was fixed
        Ver 2.72 2007-10-23  minor revision to robust against the header corrupted data  
        Ver 2.73 2008-02-19  options '--help' and '--ver[sion]' were added 
        Ver 2.74 2009-07-31  just reflect the version up of libvssp32.c
        Ver 2.75 2009-10-15  minor bug in "chkfilen" was fixed
        Ver 2.76 2011-02-07  all "long int" were changed to "int"

  NOTE: Use of compiler option "-O3" will result in error at inline-assembler
        function. Use "-O2" option.


D:\IPVLBI\data\errordata\test02.dat 0 -test.log   ---  VSSP32 defective data
D:\IPVLBI\data\errordata\S102084410b.dat.old --- VSSP slip data
D:\IPVLBI\data\errordata\errorGSIsample.k5 --- VSSP32 header error data
 
*/
#define PROGRAM_NAME "datachk"
#define PROGRAM_VERSION "Ver. 2011-02-07"

#include "libk5.h"      // common header for IP-VLBI utilities
#include "libvssp32.h"

//#ifndef WIN32
//#include <stdint.h>
//#endif
//#define REWIND_BYTE  0  // define how many bytes rewind if failure in 
                          // header detection  on 1st trial (shoud be multiples of 4)
                          // 0 means rewind till previous header
#ifdef WIN32
  #ifndef USE_ASM
    #define USE_ASM    // use inline assemble function
  #endif
#endif

#define SYNC_SEARCH_BYTE_STEP  // Sync pattern search step
//#define SYNC_SEARCH_2BYTES_STEP 
//#define SYNC_SEARCH_4BYTES_STEP 


#define COUNTER_FILE_DATACHK "counter_file_datachk.tmp"   // counter file name 

//#define REWIND_BYTE 200 
//#define USE_ASM    // use inline assemble function

typedef struct DATACHK_INTERNAL_USE  // for datachk internal use
{
    char fname[128], keepfname[128];
    int nodatfile;  // data file not found flag  1: not found 0: found
    int mode; // processing mode
    int bitslip_by_frame[3600];
    int total_frame;
    int err_all;  // # of frame where any error occurred
    //int err_frame;  // # of frame where data are substituted by artifical data (VSSP32 only)
    int eflg_frame;  // # of frame where data are substituted by artifical data (VSSP32 only)
    //int bad_frame;  // # of bit slip or bit make occurred frame
    int slip_frame;  // # of bit slip or bit make occurred frame
    int aux_err;     // # of aux field error frame
    int aux_misalign;     // # of aux field misalign frame
    int bit_rev_err;     // # of 1 bit reversal occurred in sync field
    int time_discon;   //# of time discontinuity  (not bit slip and make)
    int time_discon_bs;  // # of time discon and bit slip or make
    int flg_1st64_ok;   // flag  1: 1st 64bit data is header  0: no 
    int byteoffset_1st_header;   // # of byte offset to detect 1st header
    int sfreq;   //sampling frequency (in kHz)
    int adbit;   // AD bit length
    int numch;   // # of CH
    int filter;  // LPF filter frequency(MHz) for K5/VSSP32 (0: through)
    int hh, mm, ss;
    int seconds; // second from 0h UTC
    int hhe, mme, sse;
    int seconds_end; // second from 0h UTC
    // following is valid for K5/VSSP32
    int year;    // 3 digit year
    int totalday;    // total day
    int yeare;    // 3 digit year
    int totaldaye;    // total day
    //
    int k5format;  // 1: K5/VSSP    2: K5/VSSP32
    int keepstatus; // 0: nothing to do  1: keep renamed  2: keep copied 
    int64_t size;  // opened file size in bytes **K5**
    int64_t tbytes;   // total # of bytes read **K5**
    char fmt[10];
} DATACHK_INTERNAL_USE_T;


void dt_intuse_ini(DATACHK_INTERNAL_USE_T *dt)
{
/*
   initialize DATACHK_INTERNAL_USE_T type variable

  */
    int n;
    dt->fname[0]='\0';
    dt->keepfname[0]='\0';
    dt->keepstatus=0;
    for(n=0; n<3600; n++) dt->bitslip_by_frame[n]=0;
    dt->total_frame=0;
    dt->err_all=0;
    dt->eflg_frame=0;
    dt->slip_frame=0;
    dt->aux_err=0;
    dt->aux_misalign=0;
    dt->bit_rev_err=0;
    dt->time_discon=0;
    dt->time_discon_bs=0;
    dt->byteoffset_1st_header=0;
    dt->sfreq=0;
    dt->adbit=0;
    dt->numch=0;
    dt->filter=0;
    dt->hh=0;
    dt->mm=0;
    dt->ss=0;
    dt->seconds=0;
    dt->hhe=0;
    dt->mme=0;
    dt->sse=0;
    dt->seconds_end=0;
    dt->year=0;
    dt->totalday=0;
    dt->yeare=0;
    dt->totaldaye=0;
    dt->fmt[0]='\0';
}

void get_counter(char* fname, int *numofile)  
// original is counterget in libcor.c
{
    FILE* stream;
    const int nbyte=128;
    char * ibuf;
    int nb;
    int i,ierr;
    ibuf=(char *)malloc(nbyte);
    //int prt[5];
    //initialize
    ierr=0;
    *numofile=0;
    if( (stream  = fopen( fname, "r+" )) == NULL )
    {
        printf("Counter File Not Found! We create file(%s)\n",fname);
        /* File creation */
        if( (stream  = fopen( fname, "w" )) == NULL ){
            printf("File Creation Error! \n");       
            goto exitend;
        }
        /* file counter initial value set */
        fprintf(stream,"0\n");
        i=fclose(stream);
        /* open again*/
        if( (stream  = fopen( fname, "r+" )) == NULL ){
            perror("File open error \n");
            goto exitend;
        }

    }

    if( fgets( ibuf, nbyte, stream ) == NULL)
    {
          //printf("File EOF or Read Error!\n");
       goto exit2;
    }
    *numofile=atoi(ibuf);
    nb=*numofile+1;
    if(nb > 10000) {
        nb=0;
    }
    /* update data */

    rewind( stream);
    fprintf(stream,"%d\n",nb);

     exit2:
     i=fclose( stream ) ;

     exitend:
        free(ibuf);
}

#ifdef USE_ASM
    #ifdef WIN32     // in-line assembler for VC++

    static void hist_count4(int* idata, int npos, int xhalf, int* xhist)
    {
        /*
           cout # of samples that exceeds xhalf 

           idata --- array to contain decoded data
           npos  --- # of samples in idata
           xhalf --  threshold
           xhist --  count results

         count every 4 data step (suppose 4ch data)

         Register assign
           eax  --  pointer to idata
           ebx  --  work (idata value)
           ecx  --  32bit-word counter
           edx  --  npos
           edi  --  plus sign counter
           esi  --  xhalf

        */
        __asm{
            mov edx, npos
            cmp edx, 0
            jz end_func
            mov eax, idata  // start address of decoded data
            mov esi, xhalf
            xor ecx,ecx   // counter for 32bit-word number
            xor edi,edi   // counter for plus sign
    loop_head:
            mov ebx, [eax+ecx*4]
            //mov esi, xhalf
            cmp ebx, esi
            jl pass1
            //inc edi
            add edi, 1    // faster than inc
    pass1:
            add ecx,4  // next  data 
            cmp ecx, edx
            jnz loop_head
    end_func:
            mov eax, xhist
            mov [eax], edi
        }
    } 
    
    static void hist_count1(int* idata, int npos, int xhalf, int* xhist)
    {
        /*
           cout # of samples that exceeds xhalf 

           idata --- array to contain decoded data
           npos  --- # of samples in idata
           xhalf --  threshold
           xhist --  count results

         count every 1 data step (suppose 1ch data)

         Register assign
           eax  --  pointer to idata
           ebx  --  work (idata value)
           ecx  --  32bit-word counter
           edx  --  npos
           edi  --  plus sign counter
           esi  --  xhalf

        */
        __asm{
            mov edx, npos
            cmp edx, 0
            jz end_func
            mov eax, idata  // start address of decoded data
            mov esi, xhalf
            xor ecx,ecx   // counter for 32bit-word number
            xor edi,edi   // counter for plus sign
    loop_head:
            mov ebx, [eax+ecx*4]
            //mov esi, xhalf
            cmp ebx, esi
            jl pass1
            //inc edi
            add edi, 1    // faster than inc
    pass1:
            add ecx,1  // next  data 
            cmp ecx, edx
            jnz loop_head
    end_func:
            mov eax, xhist
            mov [eax], edi
        }
    }
    #else     // gcc inline assemble

    static void hist_count4(int* idata, int npos, int xhalf, int* xhist)
    {
        /*
           idata --- array to contain decoded data   8(%ebp)
           npos  --- # of 32 bits word data in idata 12(%ebp)
           xhist --  array to contain results

           cout # of samples that exceeds xhalf 

           idata --- array to contain decoded data 8(%ebp)
           npos  --- # of samples in idata        12(%ebp)
           xhalf --  threshold                    16(%ebp)
           xhist --  count results                20(%ebp)

         count every 4 data step (suppose 4ch data)

         Register assign
           eax  --  pointer to idata
           ebx  --  work (idata value)
           ecx  --  32bit-word counter
           edx  --  npos
           edi  --  plus sign counter
           esi  --  xhalf

        */
        asm volatile (
            "pushl %eax \n\t"     // save register used in this function
            "pushl %ebx \n\t"
            "pushl %ecx \n\t"
            "pushl %edx \n\t"
            "pushl %edi \n\t"
            "pushl %esi \n\t"
            "movl 12(%ebp), %edx \n\t"
            "cmp $0, %edx  \n\t"
            "jz end_func4 \n\t"
            "movl 8(%ebp), %eax  \n\t" // start address of decoded data
            "movl 16(%ebp), %esi \n\t"
            "xor %ecx, %ecx \n\t"   // counter for 32bit-word number
            "xor %edi, %edi \n\t"   // counter for plus sign
    "loop_head4: \n\t"
            "movl (%eax,%ecx,4), %ebx \n\t"
            "cmp %esi, %ebx  \n\t"
            "jl pass14 \n\t"
            "add $1, %edi \n\t"    // faster than inc
    "pass14: \n\t"
            "add $4, %ecx  \n\t"  // next  data 
            "cmp %edx, %ecx \n\t"
            "jnz loop_head4 \n\t"
    "end_func4: \n\t"
            "movl 20(%ebp), %eax  \n\t"
            "movl %edi, (%eax) \n\t"
            "popl %esi \n\t"    // recover saved register 
            "popl %edi \n\t"
            "popl %edx \n\t"
            "popl %ecx \n\t"
            "popl %ebx \n\t"
            "popl %eax \n\t"
        );
    } 


    static void hist_count1(int* idata, int npos, int xhalf, int* xhist)
    {
        /*
           idata --- array to contain decoded data   8(%ebp)
           npos  --- # of 32 bits word data in idata 12(%ebp)
           xhist --  array to contain results

           cout # of samples that exceeds xhalf 

           idata --- array to contain decoded data 8(%ebp)
           npos  --- # of samples in idata        12(%ebp)
           xhalf --  threshold                    16(%ebp)
           xhist --  count results                20(%ebp)

         count every 1 data step (suppose 1ch data)

         Register assign
           eax  --  pointer to idata
           ebx  --  work (idata value)
           ecx  --  32bit-word counter
           edx  --  npos
           edi  --  plus sign counter
           esi  --  xhalf

        */
        asm volatile (
            "pushl %eax \n\t"     // save register used in this function
            "pushl %ebx \n\t"
            "pushl %ecx \n\t"
            "pushl %edx \n\t"
            "pushl %edi \n\t"
            "pushl %esi \n\t"
            "movl 12(%ebp), %edx \n\t"
            "cmp $0, %edx  \n\t"
            "jz end_func1 \n\t"
            "movl 8(%ebp), %eax  \n\t" // start address of decoded data
            "movl 16(%ebp), %esi \n\t"
            "xor %ecx, %ecx \n\t"   // counter for 32bit-word number
            "xor %edi, %edi \n\t"   // counter for plus sign
    "loop_head1: \n\t"
            "movl (%eax,%ecx,4), %ebx \n\t"
            "cmp %esi, %ebx  \n\t"
            "jl pass11 \n\t"
            "add $1, %edi \n\t"    // faster than inc
    "pass11: \n\t"
            "add $1, %ecx  \n\t"  // next  data 
            "cmp %edx, %ecx \n\t"
            "jnz loop_head1 \n\t"
    "end_func1: \n\t"
            "movl 20(%ebp), %eax  \n\t"
            "movl %edi, (%eax) \n\t"
            "popl %esi \n\t"    // recover saved register 
            "popl %edi \n\t"
            "popl %edx \n\t"
            "popl %ecx \n\t"
            "popl %ebx \n\t"
            "popl %eax \n\t"
        );
    } 

    #endif

#else

    void hist_count4(int* idata, int npos, int xhalf, int* pcount)
    {
        int n,k;
        *pcount=0;
        for(n=0; n<npos; n+=4){
           if(idata[n] >= xhalf) *pcount+=1;
        }
    }

    void hist_count1(int* idata, int npos, int xhalf, int* pcount)
    {
        int n;
        *pcount=0;
        for(n=0; n<npos; n++){
           if(idata[n] >= xhalf) *pcount+=1;
        }
    }

#endif

void hist_count(K5VSSP_file_t *K5, int *xhist)
{
    /*  do count for statistics */
    int xhistw[8];     // 8=4ch*2  (" means sign+  sign- of data)
    int xhalf,n;
    xhalf=1;
    xhalf<<=K5->adbit-1;
    if(K5->numch==4){  // case of 4 channel data
         hist_count4(K5->idata, K5->npos, xhalf, &xhistw[1]);
         hist_count4(K5->idata+1, K5->npos, xhalf, &xhistw[3]);
         hist_count4(K5->idata+2, K5->npos, xhalf, &xhistw[5]);
         hist_count4(K5->idata+3, K5->npos, xhalf, &xhistw[7]);
         xhistw[0]=K5->npos/K5->numch-xhistw[1];
         xhistw[2]=K5->npos/K5->numch-xhistw[3];
         xhistw[4]=K5->npos/K5->numch-xhistw[5];
         xhistw[6]=K5->npos/K5->numch-xhistw[7];
         for(n=0; n<8; n++){
             xhist[n]+=xhistw[n];
         }
    } else {  // case of 1 ch data
         hist_count1(K5->idata, K5->npos, xhalf, &xhistw[1]);
         xhistw[0]=K5->npos-xhistw[1];
         xhist[0]+=xhistw[0];
         xhist[1]+=xhistw[1];
    }
}

void hist_count_full(K5VSSP_file_t *K5, int *xhist)
{
    /*  
        count for statistics 
        with full reso;ution

        xhist[4*256]  -- occurrence by ch, by amp 

    */
    int n,k;
    //printf("Debug K5->npos  %d\n",K5->npos);
    for(n=0; n<K5->numch; n++){
        for(k=0; k<K5->npos/K5->numch; k++){
            //printf("%d ",K5->idata[k*K5->numch+n]);
            xhist[n*256+K5->idata[k*K5->numch+n]]++;
        }
    }
}

void parseADdata_3N_without_check(K5VSSP_file_t *K5)
{
/*
    decode K5/VSSP (IP-VLBI) data into integer data

      [in]
         K5->ubuf    ---  sampler (K5/VSSP (IP-VLBI) board) out data 
         K5->wordread --  # of data in K5->ubuf (32bit word unit)
         K5->adbit -- sampler A/D resolution bits (1,2,4,8)
      [return]
         K5->idata -- data parsed into each sampled data
         K5->npos  -- valid number of data in xdata array

     Relation between # of ch (numch) and each ch time siries

        sdata(m,k)=xdata(k*numch+m)

             where  numch :  # of channels  (1, 4)
                      m   :  CH#
                      k   :  k-th sampling
        
*/
     int numb8, pat, n, ilast, ipos;
     int pattbl[9]={1,1,3,3,15,15,15,15,255};
     u_int32_t syncpat=0x8B000000UL;    // 2nd sync pattern for K5/VSSP
     u_int32_t syncpat32=0x8C000000UL;  // 2nd sync pattern for K5/VSSP32
     u_int32_t wd1,wd2;
     int *K5pos;
     unsigned char bpat;

     if(K5->vflag != 1) goto exitend;
     //numb8=numb;
     numb8=K5->wordread*4;

     ipos=0;
     ilast=32/K5->adbit;
     pat=pattbl[K5->adbit];
     bpat=(unsigned char)pat;
     wd1=K5->ubuf[0];

     //printf("Debug IN PARSE adbit, pat %d %d\n",K5->adbit,pat);
     K5pos=K5->idata;
     for( n=1; n < K5->wordread; n++){
        wd2=K5->ubuf[n];
        /* check sync pattern */
    //    if ( (wd1 == 0xFFFFFFFFUL) && (((wd2 >> 24) & 0xFF) == (K5->spatn & 0xFFUL)) ) { // sync detected for K5/VSSP
    //        printf("Sync detected \n"); 
    //        continue;  // pass decoding
    //    }
        /* use of inline assembler */
        K5pos=&K5->idata[ipos];
        decode_data(K5pos, wd1, ilast, pat, K5->adbit);   // use of inline assembler
        //decode_data2(K5pos, wd1, ilast, pat, K5->adbit); // normal C function
                              // no big difference found between decode_data and decode data2
                              //   ex. decode_data  34.279 sec for 8MHz 1bit 4CH 20 sec data
                              //   ex. decode_data2 34.289 sec for 8MHz 1bit 4CH 20 sec data
                              //  direct use like below is also almost same processing speed
                              //     i.e. 34.289 sec
/*
        __asm{
            mov edx, ilast
            mov eax, K5pos
            mov ebx, wd1
            mov cl, bpat
            xor esi,esi  
    loop_1:
            mov edi, ebx
            and edi, pat
            mov [eax+esi*4], edi
            shr ebx, cl
            inc esi
            cmp esi, edx
            jnz loop_1
        }
*/
        ipos+=ilast;

        wd1=wd2;
        //K5pos+=ilast;

     }
/*     
     for(i=0; i<ilast; i++){
         ww=(int)(wd1 & pat);
         //printf("ww wd1 pat %d %d %d",ww,wd1,pat); //debug
         K5->idata[ipos]=ww;
         wd1=wd1 >> K5->adbit ; 
         ipos++;
     }
*/     

     //K5pos=K5->idata+ipos;
     K5pos=&K5->idata[ipos];
     decode_data(K5pos, wd1, ilast, pat, K5->adbit);  
     //decode_data2(K5pos, wd1, ilast, pat, K5->adbit);
     /* use of inline assembler */
/*
     __asm{
        mov edx, ilast
        mov eax, K5pos
        mov ebx, ww
        mov cl, bpat
        xor esi,esi  
loop_2:
        mov edi, ebx
        and edi, pat
        mov [eax+esi*4], edi
        shr ebx, cl
        inc esi
        cmp esi, edx
        jnz loop_2
     }
*/
     ipos+=ilast;

     K5->npos=ipos;

exitend:;
     //printf("numb8,ipos %d %d\n",numb8,ipos);
}


int chkfilen3(DATACHK_INTERNAL_USE_T *dt)
{
    /* check K5 data with sampling statistics by AD level */
    /* 2007-09-23 */

       int usampl=200000;
       int i,n,nch,level;
       //int kword,kbytes;
       int kbytes;
       int64_t tword,tbytes;
       //long int numb,npos, secold, tflag, nkai;
       //long int bitsec,bitint;
       int numb,npos, secold, tflag, nkai;  // 2011-02-07
       int bitsec,bitint;                   // 2011-02-07
       //int rword;
       int f_c;
       //int xhist[8];     // 8=4ch*2 level (" means sign+  sign- of data)
       int xhist[4*256];     // 4ch*256 level   2007-09-23
       int shist[4];   //2007-09-25
       int xhalf;
       char cbuf[10], fname2[80];
       char fname[120];
       char fmt[10];
       int total_frame, slip_frame;
       //int wordread;
       int headerbits;
       double dbytes;
       K5VSSP_file_t K5;
       vsspaux_t aux;

       k5vssp_t_ini(&K5);

       total_frame=0;
       slip_frame=0;

       numb=0;
       npos=0;
       nkai=0;

       tflag=0;
       f_c=0;
       dbytes=0.0;
       //for(i=0; i<8; i++) xhist[i]=0; 
       for(i=0; i<4*256; i++) xhist[i]=0; // 2007-09-23
       for(i=0; i<4; i++) shist[i]=0; // 2007-09-25


       /* K5/VSSP file open and 1st header read 
               to get sampling information etc */

       strcpy(fname,dt->fname);
       strcpy(K5.fname,fname);

       dt->nodatfile=0;
       K5.size=get_file_size(K5.fname);
       if(K5.size == -1){
           printf("File %s not found! \n",K5.fname);
           dt->nodatfile=1;
           goto exitend2;
       }
       dt->size=K5.size;
       if( (K5.stream  = fopen( K5.fname, "rb" )) == NULL ) {
           printf("File %s Open Error! \n",K5.fname);
           goto exitend2;
       }
       printf("FILE : %s  (%lld bytes)\n",K5.fname,K5.size);
       //printf("FILE : %s  (%u bytes)\n",K5.fname,K5.size);

       tword=0;
       tbytes=0;

       //if(checkheader(&K5) < 0){
       if((kbytes=CheckK5header_with_Read(&K5)) < 0){
            printf("File read error! \n");
            //fgetpos(K5.stream,&K5.pos); 
            i=fclose( K5.stream ) ;
            goto exitend2;
       }
       if(kbytes >0){
           printf("First 64bit is not a header block!\n");
       }

       if(K5.format==2) headerbits=96+K5.auxbyte*8;  // usually 256
       //printf("headerbits= %d\n",headerbits);


       dbytes=kbytes*1.0+K5.headsizew*4.0;
       //printf("kword, headsizew, dbytes %d %d %f\n",kword,K5.headsizew,dbytes);

       //tword+=K5.twordread;
       tbytes+=K5.tbytesread;
       K5.twordread=0;
       K5.tbytesread=0;

       K5.fsample=1000.0*K5.sfreq;   // sampling frequency (Hz)
       K5.tsample=1.0/K5.fsample;            // sampling period (sec)


       //bitsec=(long)K5.fsample*K5.adbit*K5.numch;  // # of bits in 1 sec data except header
       bitsec=(int)K5.fsample*K5.adbit*K5.numch;  // # of bits in 1 sec data except header  2011-02-07
       K5.bytesec=bitsec/8;                  //# of bytes in 1 sec data
       K5.imax=(K5.sfreq*1000)/usampl;        // # of usampls in a second
       K5.numb=usampl*K5.adbit*K5.numch/8;       // # of bytes in a usampl
       K5.nword=K5.numb/4;                 // # of 32bit-word in a usampl

       K5.unitp=(double)usampl*K5.tsample;   // period in a usampl (sec)

       //K5.ibuf  = (unsigned char *)malloc(K5.numb); // K5 data read buffer
       K5.ubuf  = (u_int32_t *)malloc(K5.nword*sizeof(u_int32_t)); // K5 data read buffer
       K5.idata  = (int *)malloc(32*K5.nword*sizeof(int)); // K5 data read buffer

       if(kbytes >0){
           printf("First 64bit is not a header block!\n");
       }

       strcpy(fmt,"VSSP");
       if(K5.format==2) strcpy(fmt,"VS32");

       if(K5.format==2){
           printf("\n This is K5/VSSP32 format data.\n");
           printf("\n  1st Extended Header Info is as follows.\n");
           printf("    Year = %d    Total Day = %d\n", K5.year,K5.totalday);
           printf("    Version  Major = %d  Minor = %d    # of AUX Bytes = %d\n",
               K5.version1,K5.version2,K5.auxbyte);
           printf("    AUX Field Data : ");
           for(n=0; n<K5.auxbyte/4; n++){
               printf(" %04X", K5.aux32[n]);
           }
           aux.size=K5.auxbyte;
           //strncpy(aux.bdat,(char *)K5.aux32,K5.auxbyte); // this is not good for in case of data 0 is included
           for(n=0; n<aux.size; n++) aux.bdat[n]=*((char *)K5.aux32+n);
           vssp_aux_parse(&aux);
           printf("\n        Format # = %d\n",aux.fmt);
           if(aux.fmt != 0)
             printf("        Filter (MHz) : %d\n",aux.filter);
           if(aux.fmt == 1)
             printf("        Station info : ID2(%s) name(%s)\n",aux.sid2,aux.station8);
           if(aux.fmt == 1 || aux.fmt == 2 )
             printf("        PC host name :  %s\n",aux.hostname8);

           printf("\n");
       }
       printf("\n");

       printf("                                                    %% of plus sign data  \n");
       printf(" FMT A/D CH  f(kHz)  TIME   seconds    # of bits    Ch1    Ch2    Ch3    Ch4\n");
       printf("%s %2d  %2d  %5d %02d:%02d:%02d %6d  ", fmt,
                  K5.adbit,K5.numch,K5.sfreq,K5.hh,K5.mm,K5.ss,K5.seconds);


       while(1){
           //printf(" FMT A/D CH  f(kHz)  TIME   seconds    # of bits\n");
           //printf("%s %2d  %2d  %5d %02d:%02d:%02d %6d  ", fmt,
           //       K5.adbit,K5.numch,K5.sfreq,K5.hh,K5.mm,K5.ss,K5.seconds);
           for(n=0; n<K5.imax; n++){
                K5.wordread = fread(K5.ubuf, sizeof(u_int32_t), K5.nword, K5.stream);
                //parseADdata_3N(&K5);
                parseADdata_3N_without_check(&K5); //2007-08-23
                dbytes+=(double)K5.wordread*4.0;
                //tword+=K5.wordread;
                tbytes+=K5.wordread*4;

                /* sampling statistics (occurrence by level) */
                hist_count_full(&K5, xhist);

                if (K5.wordread ==0 ) {// EOF
                    i=fclose( K5.stream ) ;
                    f_c++;
                    // continuous file name creation     XXXX.datN   where N=1,2,3......
                    strcpy(fname2,fname);
                    sprintf(cbuf,"%d\0",f_c);
                    strcat(fname2,cbuf);
                    /* file open */
                    if( (K5.stream  = fopen( fname2, "rb" )) == NULL ){
                        // file not found
                        goto exitend2;
                    }   
                    printf("\n continuous file (%s) opened!\n",fname2);  
                    printf("                                   ");  
                    strcpy(K5.fname,fname2);
                    K5.wordread = fread( K5.ubuf, sizeof( u_int32_t ), K5.nword, K5.stream );
                    //parseADdata_3N(&K5);
                    parseADdata_3N_without_check(&K5); //2007-08-23
                    dbytes+=(double)K5.wordread*4.0;
                    //tword+=K5.wordread;
                    tbytes+=K5.wordread*4;

                    /* sampling statistics by AD level */
                    hist_count_full(&K5, xhist);

                }
                if( ferror( K5.stream )  ) {
                    //printf("File Read Error!\n");
                    break;
                }

           }
           K5.tbytesread=0;  // reset bytes read counter
           if((kbytes=CheckK5header_with_Read(&K5)) < 0){
                tbytes+=K5.tbytesread;
                //printf("File read error! \n");
                //fgetpos(K5.stream,&K5.pos); 
                i=fclose( K5.stream ) ;
                goto exitend2;
           }
           //dbytes+=kword*4.0+K5.headsizew*4.0;
           dbytes+=kbytes*1.0+K5.headsizew*4.0;
           //tword+=K5.twordread;
           tbytes+=K5.tbytesread;
           // statistics print

           K5.vflag=1;
           xhalf=(int)(pow(2,K5.adbit-1)+0.5);
           //printf("Debug xhalf %d\n",xhalf);
           total_frame++;
           //bitint=(numb-8)*8;
           //bitint=(numb-2)*32;
           numb=kbytes;
           //printf("kword, k5.headsizew %d %d\n",kword,K5.headsizew);
           bitsec=K5.sfreq*1000*K5.adbit*K5.numch;
           bitint=bitsec+kbytes*8;
           printf("%12d\n",bitint);

           printf("\n                         # of OCCURRENCES\n");
           printf("LEVEL         CH1         CH2         CH3         CH4\n");

           for(level=0; level<(int)pow(2.0,(double)K5.adbit); level++){
               printf("%5d",level);
               for(nch=0; nch<K5.numch; nch++){
                  printf("  %10d",xhist[nch*256+level]);
                  shist[nch]+=xhist[nch*256+level];
               }
               printf("\n");
           }
           printf("  SUM");
           for(nch=0; nch<K5.numch; nch++){
              printf("  %10d",shist[nch]);
           }
           printf("\n");

           if (numb != 0){
               slip_frame++;
               printf("NG "); 
               if(tflag ==0){
                   printf("Warning! (# of bits Error)\n"); 
               } else {
                   printf("Warning! (TIME DISCON and Bits error!)\n");
               }

           } else {
               if(tflag==0){
                   printf("  \n"); 
               } else {
                   printf(" Warning! (TIME DISCON)\n");
               }
           }
           if(nkai >0 && K5.seconds-secold !=1){
                printf("%s %2d  %2d  %5d %02d:%02d:%02d %6dNG", fmt,
                          K5.adbit,K5.numch,K5.sfreq,K5.hh,K5.mm,K5.ss,K5.seconds);
                tflag=1;
           } else {
                printf("%s %2d  %2d  %5d %02d:%02d:%02d %6d  ", fmt,
                          K5.adbit,K5.numch,K5.sfreq,K5.hh,K5.mm,K5.ss,K5.seconds);
                tflag=0;
           }
           for(i=0; i<4*256; i++) xhist[i]=0; //2007-09-23
           for(i=0; i<4; i++) shist[i]=0; // 2007-09-25
           secold=K5.seconds;
           nkai++;
       }

 exitend2:
        if(dt->nodatfile == 1) return -1;

        free(K5.ubuf);
        free(K5.idata);

        dt->tbytes=tbytes;

        printf("\n");
        printf("%lld bytes read\n",tbytes); 
        printf("%d frames out of %d frames (%f%%) are bad frames.\n",
           slip_frame, total_frame, 
           100.0*(double)slip_frame/(double)total_frame);
       return slip_frame;
}



int chkfilen2(DATACHK_INTERNAL_USE_T *dt)
{
    /* check K5 data with sampling statistics */

       int usampl=200000;
       int i,n;
       //int kword,kbytes;
       int kbytes;
       int64_t tword,tbytes;
       //long int numb,npos, secold, tflag, nkai;
       //long int bitsec,bitint;
       int numb,npos, secold, tflag, nkai;  // 2011-02-7
       int bitsec,bitint;                   // 2011-02-7
       //int rword;
       int f_c;
       int xhist[8];     // 8=4ch*2 level (" means sign+  sign- of data)
       int xhalf;
       char cbuf[10], fname2[80];
       char fname[120];
       char fmt[10];
       int total_frame, slip_frame;
       //int wordread;
       int headerbits;
       double dbytes;
       K5VSSP_file_t K5;
       vsspaux_t aux;

       k5vssp_t_ini(&K5);

       total_frame=0;
       slip_frame=0;

       numb=0;
       npos=0;
       nkai=0;

       tflag=0;
       f_c=0;
       dbytes=0.0;
       for(i=0; i<8; i++) xhist[i]=0; 


       /* K5/VSSP file open and 1st header read 
               to get sampling information etc */

       strcpy(fname,dt->fname);
       strcpy(K5.fname,fname);

       dt->nodatfile=0;
       K5.size=get_file_size(K5.fname);
       if(K5.size == -1){
           printf("File %s not found! \n",K5.fname);
           dt->nodatfile=1;
           goto exitend2;
       }
       dt->size=K5.size;
       if( (K5.stream  = fopen( K5.fname, "rb" )) == NULL ) {
           printf("File %s Open Error! \n",K5.fname);
           goto exitend2;
       }
       printf("FILE : %s  (%lld bytes)\n",K5.fname,K5.size);
       //printf("FILE : %s  (%u bytes)\n",K5.fname,K5.size);

       tword=0;
       tbytes=0;

       //if(checkheader(&K5) < 0){
       if((kbytes=CheckK5header_with_Read(&K5)) < 0){
            printf("File read error! \n");
            //fgetpos(K5.stream,&K5.pos); 
            i=fclose( K5.stream ) ;
            goto exitend2;
       }
       if(kbytes >0){
           printf("First 64bit is not a header block!\n");
       }

       if(K5.format==2) headerbits=96+K5.auxbyte*8;  // usually 256
       //printf("headerbits= %d\n",headerbits);


       dbytes=kbytes*1.0+K5.headsizew*4.0;
       //printf("kword, headsizew, dbytes %d %d %f\n",kword,K5.headsizew,dbytes);

       //tword+=K5.twordread;
       tbytes+=K5.tbytesread;
       K5.twordread=0;
       K5.tbytesread=0;

       K5.fsample=1000.0*K5.sfreq;   // sampling frequency (Hz)
       K5.tsample=1.0/K5.fsample;            // sampling period (sec)


       //bitsec=(long)K5.fsample*K5.adbit*K5.numch;  // # of bits in 1 sec data except header
       bitsec=(int)K5.fsample*K5.adbit*K5.numch;  // # of bits in 1 sec data except header 2011-02-07
       K5.bytesec=bitsec/8;                  //# of bytes in 1 sec data
       K5.imax=(K5.sfreq*1000)/usampl;        // # of usampls in a second
       K5.numb=usampl*K5.adbit*K5.numch/8;       // # of bytes in a usampl
       K5.nword=K5.numb/4;                 // # of 32bit-word in a usampl

       K5.unitp=(double)usampl*K5.tsample;   // period in a usampl (sec)

       //K5.ibuf  = (unsigned char *)malloc(K5.numb); // K5 data read buffer
       K5.ubuf  = (u_int32_t *)malloc(K5.nword*sizeof(u_int32_t)); // K5 data read buffer
       K5.idata  = (int *)malloc(32*K5.nword*sizeof(int)); // K5 data read buffer

       if(kbytes >0){
           printf("First 64bit is not a header block!\n");
       }

       strcpy(fmt,"VSSP");
       if(K5.format==2) strcpy(fmt,"VS32");

       if(K5.format==2){
           printf("\n This is K5/VSSP32 format data.\n");
           printf("\n  1st Extended Header Info is as follows.\n");
           printf("    Year = %d    Total Day = %d\n", K5.year,K5.totalday);
           printf("    Version  Major = %d  Minor = %d    # of AUX Bytes = %d\n",
               K5.version1,K5.version2,K5.auxbyte);
           printf("    AUX Field Data : ");
           for(n=0; n<K5.auxbyte/4; n++){
               printf(" %04X", K5.aux32[n]);
           }
           aux.size=K5.auxbyte;
           //strncpy(aux.bdat,(char *)K5.aux32,K5.auxbyte); // this is not good for in case of data 0 is included
           for(n=0; n<aux.size; n++) aux.bdat[n]=*((char *)K5.aux32+n);
           vssp_aux_parse(&aux);
           printf("\n        Format # = %d\n",aux.fmt);
           if(aux.fmt != 0)
             printf("        Filter (MHz) : %d\n",aux.filter);
           if(aux.fmt == 1)
             printf("        Station info : ID2(%s) name(%s)\n",aux.sid2,aux.station8);
           if(aux.fmt == 1 || aux.fmt == 2 )
             printf("        PC host name :  %s\n",aux.hostname8);

           printf("\n");
       }
       printf("\n");

       printf("                                                    %% of plus sign data  \n");
       printf(" FMT A/D CH  f(kHz)  TIME   seconds    # of bits    Ch1    Ch2    Ch3    Ch4\n");
       printf("%s %2d  %2d  %5d %02d:%02d:%02d %6d  ", fmt,
                  K5.adbit,K5.numch,K5.sfreq,K5.hh,K5.mm,K5.ss,K5.seconds);


       while(1){
           for(n=0; n<K5.imax; n++){
                K5.wordread = fread(K5.ubuf, sizeof(u_int32_t), K5.nword, K5.stream);
                //parseADdata_3N(&K5);
                parseADdata_3N_without_check(&K5); //2007-08-23
                dbytes+=(double)K5.wordread*4.0;
                //tword+=K5.wordread;
                tbytes+=K5.wordread*4;

                /* sampling balance statistics (only check +/- unbalance) */
                hist_count(&K5, xhist);

                if (K5.wordread ==0 ) {// EOF
                    i=fclose( K5.stream ) ;
                    f_c++;
                    // continuous file name creation     XXXX.datN   where N=1,2,3......
                    strcpy(fname2,fname);
                    sprintf(cbuf,"%d\0",f_c);
                    strcat(fname2,cbuf);
                    /* file open */
                    if( (K5.stream  = fopen( fname2, "rb" )) == NULL ){
                        // file not found
                        goto exitend2;
                    }   
                    printf("\n continuous file (%s) opened!\n",fname2);  
                    printf("                                   ");  
                    strcpy(K5.fname,fname2);
                    K5.wordread = fread( K5.ubuf, sizeof( u_int32_t ), K5.nword, K5.stream );
                    //parseADdata_3N(&K5);
                    parseADdata_3N_without_check(&K5); //2007-08-23
                    dbytes+=(double)K5.wordread*4.0;
                    //tword+=K5.wordread;
                    tbytes+=K5.wordread*4;

                    /* sampling balance statistics (only check +/- unbalance) */
                    hist_count(&K5, xhist);

                }
                if( ferror( K5.stream )  ) {
                    //printf("File Read Error!\n");
                    break;
                }

           }
           K5.tbytesread=0;  // reset bytes read counter
           if((kbytes=CheckK5header_with_Read(&K5)) < 0){
                tbytes+=K5.tbytesread;
                //printf("File read error! \n");
                //fgetpos(K5.stream,&K5.pos); 
                i=fclose( K5.stream ) ;
                goto exitend2;
           }
           //dbytes+=kword*4.0+K5.headsizew*4.0;
           dbytes+=kbytes*1.0+K5.headsizew*4.0;
           //tword+=K5.twordread;
           tbytes+=K5.tbytesread;
           // statistics print

           K5.vflag=1;
           xhalf=(int)(pow(2,K5.adbit-1)+0.5);
           total_frame++;
           //bitint=(numb-8)*8;
           //bitint=(numb-2)*32;
           numb=kbytes;
           //printf("kword, k5.headsizew %d %d\n",kword,K5.headsizew);
           bitsec=K5.sfreq*1000*K5.adbit*K5.numch;
           bitint=bitsec+kbytes*8;
           if(K5.numch==1){
               printf("%12d %6.1f",bitint,
                      100.0*(double)xhist[1]/(double)(xhist[0]+xhist[1]));
           } else {
               printf("%12d %6.1f %6.1f %6.1f %6.1f",bitint,
                      100.0*(double)xhist[1]/(double)(xhist[0]+xhist[1]),
                      100.0*(double)xhist[1*2+1]/(double)(xhist[1*2+0]+xhist[1*2+1]),
                      100.0*(double)xhist[2*2+1]/(double)(xhist[2*2+0]+xhist[2*2+1]),
                      100.0*(double)xhist[3*2+1]/(double)(xhist[3*2+0]+xhist[3*2+1]));
           } 
           if (numb != 0){
               slip_frame++;
               printf("NG "); 
               if(tflag ==0){
                   printf("Warning! (# of bits Error)\n"); 
               } else {
                   printf("Warning! (TIME DISCON and Bits error!)\n");
               }

           } else {
               if(tflag==0){
                   printf("  \n"); 
               } else {
                   printf(" Warning! (TIME DISCON)\n");
               }
           }
           if(nkai >0 && K5.seconds-secold !=1){
                printf("%s %2d  %2d  %5d %02d:%02d:%02d %6dNG", fmt,
                          K5.adbit,K5.numch,K5.sfreq,K5.hh,K5.mm,K5.ss,K5.seconds);
                tflag=1;
           } else {
                printf("%s %2d  %2d  %5d %02d:%02d:%02d %6d  ", fmt,
                          K5.adbit,K5.numch,K5.sfreq,K5.hh,K5.mm,K5.ss,K5.seconds);
                tflag=0;
           }
           for(i=0; i<8; i++) xhist[i]=0; 
           secold=K5.seconds;
           nkai++;
       }

 exitend2:
        if(dt->nodatfile == 1) return -1;

        free(K5.ubuf);
        free(K5.idata);

        dt->tbytes=tbytes;

        printf("\n");
        printf("%lld bytes read\n",tbytes); 
        printf("%d frames out of %d frames (%f%%) are bad frames.\n",
           slip_frame, total_frame, 
           100.0*(double)slip_frame/(double)total_frame);
       return slip_frame;
}


int chkfilen(DATACHK_INTERNAL_USE_T *dt) 
{
    /* check K5 data (headder position only) */

       int usampl=200000;
       //int i,n,kword, kbytes;
       int i,n,kbytes;
       int64_t tword, tbytes;
       //long int numb,npos, secold, tflag, nkai;
       //long int bitsec,bitint;
       int numb,npos, secold, tflag, nkai;  // 2011-02-07
       int bitsec,bitint;                   // 2011-02-07
       int headerbits;
       int f_c;
       int total_frame, slip_frame, time_discon,time_discon_bs,aux_err;
       int bit_rev_err, aux_misalign;
       int err_all;
       int eflg_frame; // statistics for VSSP32 only
       char cbuf[10], fname2[80];
       char fname[120];
       char fmt[10];
       //char lbuf[120],wbuf[60];
       char lbuf[200],wbuf[60]; //2009-10-15
       int print_flag;  // line print flag 0: no print   1: print
       double dbytes;
       K5VSSP_file_t K5;
       vsspaux_t aux;
       //fpos_t size;

       k5vssp_t_ini(&K5);
       total_frame=0;
       eflg_frame=0; // statistics for VSSP32 error flag only
       slip_frame=0;
       aux_err=0;
       aux_misalign=0;
       bit_rev_err=0;
       time_discon=0;
       time_discon_bs=0;
       numb=0;
       npos=0;
       nkai=0;

       tflag=0;
       f_c=0;
       dbytes=0.0;

       headerbits=64;  // for K5/VSSP
       lbuf[0]='\0';

       strcpy(fname,dt->fname);
       strcpy(K5.fname,fname);

       dt->flg_1st64_ok=1;   // flag default: 1st 64bits of file is headder  
       dt->filter=-1;


       /* K5/VSSP file open and 1st header read 
               to get sampling information etc */

       dt->nodatfile=0;
       K5.size=get_file_size(K5.fname);
       if(K5.size == -1){
           printf("File %s not found! \n",K5.fname);
           dt->nodatfile=1;
           goto exitend2;
       }
       dt->size=K5.size;
       if( (K5.stream  = fopen( K5.fname, "rb" )) == NULL ) {
           //printf("File %s Open Error! \n",K5.fname);
           printf("File %s not found! \n",K5.fname);
           goto exitend2;
       }
       printf("FILE : %s  (%lld bytes)\n",K5.fname,K5.size);
       //printf("FILE : %s  (%u bytes)\n",K5.fname,K5.size);

       tword=0;
       tbytes=0;

       //if(checkheader(&K5) < 0){
       if((kbytes=CheckK5header_with_Read(&K5)) < 0){
       //if((kbytes=CheckK5header_with_Read_but_no_search(&K5)) < 0){  // debug
            printf("File read error! \n");
            //fgetpos(K5.stream,&K5.pos); 
            i=fclose( K5.stream ) ;
            goto exitend2;
       }

       dt->k5format=K5.format;

       if(K5.format==2) headerbits=96+K5.auxbyte*8;  // usually 256
       //printf("headerbits= %d\n",headerbits);

       tbytes+=K5.tbytesread;
       K5.twordread=0;
       K5.tbytesread=0;

       K5.fsample=1000.0*K5.sfreq;   // sampling frequency (Hz)
       K5.tsample=1.0/K5.fsample;            // sampling period (sec)


       //bitsec=(long)K5.fsample*K5.adbit*K5.numch;  // # of bits in 1 sec data except header
       bitsec=(int)K5.fsample*K5.adbit*K5.numch;  // # of bits in 1 sec data except header 2011-02-07
       K5.bytesec=bitsec/8;                  //# of bytes in 1 sec data
       K5.imax=(K5.sfreq*1000)/usampl;        // # of usampls in a second
       K5.numb=usampl*K5.adbit*K5.numch/8;       // # of bytes in a usampl
       K5.nword=K5.numb/4;                 // # of 32bit-word in a usampl

       K5.unitp=(double)usampl*K5.tsample;   // period in a usampl (sec)

       //K5.ibuf  = (unsigned char *)malloc(K5.numb); // K5 data read buffer
       K5.ubuf  = (u_int32_t *)malloc(K5.nword*sizeof(u_int32_t)); // K5 data read buffer

       dt->byteoffset_1st_header=kbytes;
       if(kbytes >0){
           printf("First 64bit is not a header block!\n");
           dt->flg_1st64_ok=0;
       }

       strcpy(fmt,"VSSP");
       if(K5.format==2) strcpy(fmt,"VS32");

       strcpy(dt->fmt,fmt);

       /* keep first frame info */
       dt->adbit=K5.adbit;
       dt->numch=K5.numch;
       dt->sfreq=K5.sfreq;
       dt->hh=K5.hh;
       dt->mm=K5.mm;
       dt->ss=K5.ss;
       dt->seconds=K5.seconds;
       if(K5.format==2){
          dt->year=K5.year;
          dt->totalday=K5.totalday;
       }

       if(K5.format==2){
           printf("\n This is K5/VSSP32 format data.\n");
           printf("\n  1st Extended Header Info is as follows.\n");
           printf("    Year = %d    Total Day = %d\n", K5.year,K5.totalday);
           printf("    Version  Major = %d  Minor = %d    # of AUX Bytes = %d\n",
               K5.version1,K5.version2,K5.auxbyte);
           printf("    AUX Field Data : ");
           for(n=0; n<K5.auxbyte/4; n++){
               printf(" %04X", K5.aux32[n]);
           }
           aux.size=K5.auxbyte;
           //strncpy(aux.bdat,(char *)K5.aux32,K5.auxbyte); // this is not good for in case of data 0 is included
           for(n=0; n<aux.size; n++) aux.bdat[n]=*((char *)K5.aux32+n);
           vssp_aux_parse(&aux);
           printf("\n        Format # = %d\n",aux.fmt);
           if(aux.fmt != 0){
             printf("        Filter (MHz) : %d\n",aux.filter);
             dt->filter=aux.filter;
           }
           if(aux.fmt == 1)
             printf("        Station info : ID2(%s) name(%s)\n",aux.sid2,aux.station8);
           if(aux.fmt == 1 || aux.fmt == 2 )
             printf("        PC host name :  %s\n",aux.hostname8);
           printf("\n");

       }
       printf("\n");

       printf(" FMT A/D CH  f(kHz)  TIME   seconds   # of bits\n");
       //printf("%s %2d  %2d   %4d %02d:%02d:%02d %6d  ", fmt,
       //           K5.adbit,K5.numch,K5.sfreq,K5.hh,K5.mm,K5.ss,K5.seconds);
       sprintf(lbuf,"%s %2d  %2d   %4d %02d:%02d:%02d %6d  \0", fmt,
                  K5.adbit,K5.numch,K5.sfreq,K5.hh,K5.mm,K5.ss,K5.seconds);

       print_flag=1;

loop1:

       if((kbytes=fwd1secK5(&K5))<0){
           tbytes+=K5.tbytesread;
           //printf("File EOF  %d\n", K5.tbytesread);
           goto exit2next;
       }
       // update bitsec data (add 2007-10-23)
       K5.fsample=1000.0*K5.sfreq;   // sampling frequency (Hz)
       K5.tsample=1.0/K5.fsample;            // sampling period (sec)
       //bitsec=(long)K5.fsample*K5.adbit*K5.numch;  // # of bits in 1 sec data except header
       bitsec=(int)K5.fsample*K5.adbit*K5.numch;  // # of bits in 1 sec data except header 2011-02-07
       // add end

       bitint=(int)(K5.tbytesread*8-headerbits);  //64bit is headder
       tbytes+=K5.tbytesread;
       K5.tbytesread=0;  // reset read counter
       total_frame++;

       sprintf(wbuf,"  %10d\0",bitint); strcat(lbuf,wbuf);
       if(K5.ierr < 0){
           if(abs(K5.ierr)/10 == 1){
              print_flag=1;
              sprintf(wbuf," 1bit reversal in sync but OK! \0"); strcat(lbuf,wbuf);
              bit_rev_err++;
           }
           if(abs(K5.ierr)%10 == 2){   // if(K5.ierr==-2){
               print_flag=1;
               sprintf(wbuf," AUX FIELD ERROR! \0"); strcat(lbuf,wbuf);
               aux_err++;
           }
           if(abs(K5.ierr)%10 == 1){   // if(K5.ierr==-1){
               print_flag=1;
               sprintf(wbuf," AUX MISALIGN! \0"); strcat(lbuf,wbuf);
               aux_misalign++;
           }
       } else {
           if(K5.eflg == 1) {
               print_flag=1;
               sprintf(wbuf," EFLG error! \0"); strcat(lbuf,wbuf);
               eflg_frame++;  // because eflg is in AUX field
           }
       }
       if (bitint != bitsec){
          print_flag=1;
          dt->bitslip_by_frame[slip_frame]=bitsec-bitint;
          slip_frame++;
          sprintf(wbuf," NG \0");  strcat(lbuf,wbuf);
          if(tflag ==0){
              sprintf(wbuf,"# of bits Error (%d)\0",bitsec-bitint); 
          } else {
              sprintf(wbuf,"TIME DISCON and Bits error!\0");
              time_discon_bs++;
          }
          strcat(lbuf,wbuf);
       } else {
          if(tflag==0){
              //printf("  \n"); 
          } else {
              print_flag=1;
              sprintf(wbuf," TIME DISCON\0"); strcat(lbuf,wbuf);
              time_discon++;
          }
       }
       if(print_flag==1) printf("%s\n",lbuf);
       if(dt->mode==0) print_flag=0;
       sprintf(lbuf,"%s %2d  %2d   %4d %02d:%02d:%02d %6d  \0", fmt, 
                  K5.adbit,K5.numch,K5.sfreq,K5.hh,K5.mm,K5.ss,K5.seconds);

       if(total_frame==1){
           if(dt->mode==0) printf("    checking.....\n");
       }
       dt->hhe=K5.hh;
       dt->mme=K5.mm;
       dt->sse=K5.ss;
       dt->seconds_end=K5.seconds;
       if(K5.format==2){
           dt->yeare=K5.year;
           dt->totaldaye=K5.totalday;
       }

       if(nkai >0 && K5.seconds-secold !=1){
           tflag=1;
           if(K5.seconds == 0 && secold==86399) tflag=0;
       } else {
           tflag=0;
       }
       secold=K5.seconds;
       nkai++;

       goto loop1;

exit2next: // next file candidate read

       i=fclose( K5.stream ) ;
       f_c++;
       // continuous file name creation     XXXX.datN   where N=1,2,3......
       strcpy(fname2,fname);
       sprintf(cbuf,"%d\0",f_c);
       strcat(fname2,cbuf);
       strcpy(K5.fname,fname2);
       K5.size=get_file_size(K5.fname);
       if(K5.size == -1){
           //printf("File %s not found! \n",K5.fname);
           goto exitend2;
       }
       dt->size+=K5.size;
       // file open 
       if( (K5.stream  = fopen( K5.fname, "rb" )) == NULL ) {
           //printf("File %s Open Error! \n",K5.fname);
           goto exitend2;
       }
       K5.pos=0;
       tbytes-=K5.tbytesread;   //  to compensate too much count
       //K5.size=get_file_size(K5.stream);

       //printf("\n continuous file (%s (%u bytes)) opened!\n",K5.fname,K5.size);  
       printf("\n continuous file (%s (%lld bytes)) opened!\n",K5.fname,K5.size);  
       printf("                                    ");  
       if((kbytes=CheckK5header_with_Read(&K5)) < 0){
            tbytes+=K5.tbytesread;
            goto exitend;
       }

       bitint=(int)(K5.tbytesread*8-headerbits);  //64bit is headder
       tbytes+=K5.tbytesread;
       K5.tbytesread=0;  // reset read counter
       total_frame++;
       sprintf(wbuf,"  %10d\0",bitint); strcat(lbuf,wbuf);
       if(K5.ierr < 0){
           if(abs(K5.ierr)/10 == 1){
              print_flag=1;
              sprintf(wbuf," 1bit reversal in sync but OK! \0"); strcat(lbuf,wbuf);
              bit_rev_err++;
           }
           if(abs(K5.ierr)%10 == 2){   // if(K5.ierr==-2){
               print_flag=1;
               sprintf(wbuf," AUX FIELD ERROR! \0"); strcat(lbuf,wbuf);
               aux_err++;
           }
           if(abs(K5.ierr)%10 == 1){   // if(K5.ierr==-1){
               print_flag=1;
               sprintf(wbuf," AUX MISALIGN! \0"); strcat(lbuf,wbuf);
               aux_misalign++;
           }
       } else {
           if(K5.eflg == 1) {
               print_flag=1;
               sprintf(wbuf," EFLG error! \0"); strcat(lbuf,wbuf);
               eflg_frame++;  // because eflg is in AUX field
           }
       }
       if (bitint != bitsec){
          print_flag=1;
          dt->bitslip_by_frame[slip_frame]=bitsec-bitint;
          slip_frame++;
          sprintf(wbuf," NG \0");  strcat(lbuf,wbuf);
          if(tflag ==0){
              sprintf(wbuf,"# of bits Error (%d)\0",bitsec-bitint); 
          } else {
              sprintf(wbuf,"TIME DISCON and Bits error!\0");
              time_discon_bs++;
          }
          strcat(lbuf,wbuf);
       } else {
          if(tflag==0){
              //printf("  \n"); 
          } else {
              print_flag=1;
              sprintf(wbuf," TIME DISCON\0"); strcat(lbuf,wbuf);
              time_discon++;
          }
       }
       if(print_flag==1) printf("%s\n",lbuf);
       if(dt->mode==0) print_flag=0;

       sprintf(lbuf,"%s %2d  %2d   %4d %02d:%02d:%02d %6d  \0", fmt,
                  K5.adbit,K5.numch,K5.sfreq,K5.hh,K5.mm,K5.ss,K5.seconds);
       dt->hhe=K5.hh;
       dt->mme=K5.mm;
       dt->sse=K5.ss;
       dt->seconds_end=K5.seconds;
       if(nkai >0 && K5.seconds-secold !=1){
           tflag=1;
           if(K5.seconds == 0 && secold==86399) tflag=0;
       } else {
           tflag=0;
       }
       secold=K5.seconds;
       goto loop1;
    
 exitend:
       i=fclose( K5.stream ) ;
       //free(K5.ibuf);
       free(K5.ubuf);
       printf("\n");
 exitend2:;
       printf("%s\n",lbuf);

       if(dt->nodatfile == 1) return -1;


       dt->tbytes=tbytes;

       //tword+=K5.twordread;
       printf("\n");
       printf("%lld bytes read\n",tbytes);
       if(dt->flg_1st64_ok==0)
       printf(" Warning! 1st 64bits is not headder!! Offset is %d bytes.\n",dt->byteoffset_1st_header);
       printf("%5d frames out of %4d frames (%6.2f%%) are bad (bit slip) frames.\n",
           slip_frame, total_frame, 
           100.0*(double)slip_frame/(double)total_frame);
       printf("%5d frames out of %4d frames (%6.2f%%) are time discon frames.\n",
           time_discon, total_frame, 
           100.0*(double)time_discon/(double)total_frame);
       printf("%5d frames out of %4d frames (%6.2f%%) are time discon and bit slip frames.\n",
           time_discon_bs, total_frame, 
           100.0*(double)time_discon_bs/(double)total_frame);
       if(K5.format==2){
          printf("%5d frames out of %4d frames (%6.2f%%) are EFLG set frames.\n",
           eflg_frame, total_frame, 
           100.0*(double)eflg_frame/(double)total_frame);
       }
       if(aux_misalign != 0){
           printf("%5d frames out of %4d frames (%6.2f%%) are misaligned AUX field.\n",
             aux_misalign, total_frame, 
             100.0*(double)aux_misalign/(double)total_frame);
       }
       if(aux_err != 0){
           printf("%5d frames out of %4d frames (%6.2f%%) are errored AUX field.\n",
             aux_err, total_frame, 
             100.0*(double)aux_err/(double)total_frame);
       }
       if(bit_rev_err != 0){
           printf("%5d frames out of %4d frames (%6.2f%%) are 1bit reversal in sync error.\n",
             bit_rev_err, total_frame, 
             100.0*(double)bit_rev_err/(double)total_frame);
       }
       dt->slip_frame=slip_frame;
       dt->total_frame=total_frame;
       dt->time_discon=time_discon;
       dt->time_discon_bs=time_discon_bs;
       dt->aux_err=aux_err;
       dt->aux_misalign=aux_misalign;
       dt->bit_rev_err=bit_rev_err;
       dt->eflg_frame=eflg_frame;
       //err_all=slip_frame+time_discon+time_discon_bs+aux_err+eflg_frame;
       err_all=slip_frame+time_discon+time_discon_bs+aux_err+eflg_frame+bit_rev_err;
       return err_all;
}

void summary_out(int imode, char* logfile, DATACHK_INTERNAL_USE_T *dt) 
{
    // summary file out
    int n;
    FILE *fo;
    if(imode>=1){
       if(imode==1) fo = fopen( logfile, "w" );
       if(imode==2) fo = fopen( logfile, "a" );
       fprintf(fo,"# File Name:\n");
       fprintf(fo,"%s\n",dt->fname);
       if(dt->k5format==2){// VSSP32
       fprintf(fo,"# FMT  A/D CH  f(kHz)  LPF(MHz):\n");
       fprintf(fo,"  %s %2d  %2d%8d %8d\n",dt->fmt,dt->adbit,dt->numch,dt->sfreq,dt->filter);
       } else { //VSSP
       fprintf(fo,"# FMT  A/D CH  f(kHz):\n");
       fprintf(fo,"  %s %2d  %2d%8d\n",dt->fmt,dt->adbit,dt->numch,dt->sfreq);
       }
       fprintf(fo,"# Start and Last Time:\n");
       if(dt->k5format==2){ // VSSP32
          fprintf(fo,"  %04d/%03d %02d:%02d:%02d %6d  \n", dt->year,dt->totalday,dt->hh,dt->mm,dt->ss,dt->seconds);
          fprintf(fo,"  %04d/%03d %02d:%02d:%02d %6d  \n", dt->yeare,dt->totaldaye,dt->hhe,dt->mme,dt->sse,dt->seconds_end);
       } else { // VSSP
          fprintf(fo,"  %02d:%02d:%02d %6d  \n", dt->hh,dt->mm,dt->ss,dt->seconds);
          fprintf(fo,"  %02d:%02d:%02d %6d  \n", dt->hhe,dt->mme,dt->sse,dt->seconds_end);
       }
       fprintf(fo,"# Duration:\n");
       if(dt->seconds_end-dt->seconds < 0){
          fprintf(fo,"%d\n",dt->seconds_end+86400-dt->seconds+1);
       } else {
          fprintf(fo,"%d\n", dt->seconds_end-dt->seconds+1);
       }
       fprintf(fo,"# Byte offset of 1st header:\n");
       fprintf(fo,"%d\n",dt->byteoffset_1st_header);
       fprintf(fo,"# STATISTICS total bad discon discon_with_bitslip aux_err EFLG bit_rev aux_misalign:\n");
       fprintf(fo,"%d %d %d %d %d %d %d %d\n",dt->total_frame,dt->slip_frame,dt->time_discon,
              dt->time_discon_bs,dt->aux_err,dt->eflg_frame, dt->bit_rev_err, dt->aux_misalign);
       if(dt->slip_frame>0){
          fprintf(fo,"# BIT SLIP:\n");
          for(n=0; n<dt->slip_frame; n++) fprintf(fo,"%d %d\n",n+1,dt->bitslip_by_frame[n]);
       } else {
          fprintf(fo,"# NO BIT SLIP:\n");
       }
       fprintf(fo,"\n");
       fclose(fo);
       printf("Summary log file: %s\n",logfile); 
    }
}

void err_log_out(int imode, char* logfile, DATACHK_INTERNAL_USE_T *dt) 
{
    // err_log out out
    FILE *fo;
    if(dt->err_all==0) goto go2return;
    if(imode>=1){
       
       if(imode==1) fo = fopen( logfile, "w" );
       if(imode==2) fo = fopen( logfile, "a" );

       fprintf(fo,"# Errored Data File Name:\n");
       fprintf(fo,"%s\n",dt->fname);
       if(dt->k5format==2){// VSSP32
       fprintf(fo,"# FMT  A/D CH  f(kHz)  LPF(MHz):\n");
       fprintf(fo,"  %s %2d  %2d%8d %8d\n",dt->fmt,dt->adbit,dt->numch,dt->sfreq,dt->filter);
       } else { //VSSP
       fprintf(fo,"# FMT  A/D CH  f(kHz):\n");
       fprintf(fo,"  %s %2d  %2d%8d\n",dt->fmt,dt->adbit,dt->numch,dt->sfreq);
       }
       if(dt->keepstatus==1){
           fprintf(fo,"# Renamed to:\n");
       } else if (dt->keepstatus==2){
           fprintf(fo,"# Copied to:\n");
       }
       fprintf(fo,"%s\n",dt->keepfname);
       fprintf(fo,"\n");
       fclose(fo);
       printf("Error log file: %s\n",logfile); 
    }
go2return:;
}

int main(int argc, char *argv[])
{
    char logname1[128], logname2[128];
    char cbuf[256];
    char* dflt_fname="tds.data";
    char* f_counter=COUNTER_FILE_DATACHK;  //serial file number counter file
    int n,mode,imode,lmode,len,i;
    int keepmode;
    int err_all;
    int number;
    double telapsed;           // to measure processing time
    clock_t cstart1sec, cfinish1sec;  // to calculate processing time
    DATACHK_INTERNAL_USE_T dt;


    mode=0;
    if(argc == 1 || (argc == 2 && strncmp(*(argv+1),"--help",6)==0 )){ // 2008-02-19
       //strcpy(fname,dflt_fname);
       printf("%s %s\n",PROGRAM_NAME,PROGRAM_VERSION);
       printf("How to run: \n");
       printf("  datachk datafile [mode [logfile [errlog [keepmode]]]]\n");
       printf("\n");
       printf("    where  datafile ---- K5 data file name to check\n");
       //printf("                          if 0 is given, internal dflt name will be used\n");
       printf("           mode     ---- 0: Just check # of bits and time (default)\n");
       printf("                           only error occured frame is diplayed.\n");
       printf("                         1: Display sampling statistics (zero balance)\n");
       printf("                         2: same as mode 0 but each frame is displayed\n");
       printf("                         3: Display sampling statistics by AD level\n");
       printf("           logfile  ---- summary out file name for mode 0 (default: no out)\n");
       printf("                          file name start with '-' means append mode open\n");
       printf("           errlog   ---- log file name to contain error \n");
       printf("                         detected data file name for mode 0.\n");
       printf("                         This file will not be created in case \n");
       printf("                          of no error\n");
       printf("                           (default: no log)\n");
       printf("                          file name start with '-' means append mode open\n");
       printf("           keepmode ---- to keep data file error occurred for mode 0.\n");
       printf("                           0: nothing to do (default)\n");
       printf("                           1: rename to \"original+.NNNN.err\"\n");
       printf("                           2: copy to \"original+.NNNN.err\"\n");
       printf("                             where NNNN is 0001 to 9999  \n");
       exit(0);
    }
    if(argc == 2 && strncmp(*(argv+1),"--ver",5)==0 ){ // 2008-02-19
       printf("%s %s\n",PROGRAM_NAME,PROGRAM_VERSION);
       exit(0);
    }

    // initialize dt
    dt_intuse_ini(&dt);


    if (argc>=2){
       strcpy(dt.fname,*(argv+1));
    } 
    if(argc >=3){
       mode = atoi(*(argv + 2 ));
    }
    imode=0;  // no summary out
    if (argc>=4){
       imode=1; // summary out
       strcpy(logname1,*(argv+3));
       if(logname1[0]=='-') { 
           imode=2; // append mode
           len=strlen(logname1);
           for(n=0; n<len-1; n++) logname1[n]=logname1[n+1]; 
           logname1[len-1]='\0';
       }
    }
    lmode=0;  // no error log out
    if (argc>=5){
       lmode=1; // error log out
       strcpy(logname2,*(argv+4));
       if(logname2[0]=='-') { 
           lmode=2; // append mode
           len=strlen(logname2);
           for(n=0; n<len-1; n++) logname2[n]=logname2[n+1]; 
           logname2[len-1]='\0';
       }
    }
    keepmode=0;  //
    if (argc>=6){
       keepmode = atoi(*(argv + 5 ));
    }
    dt.mode=mode;
    cstart1sec = clock();
    printf("************ %s  %s ************\n",PROGRAM_NAME,PROGRAM_VERSION);
    printf("     [Note: # of bits does not include Header]\n");
    if(mode==3){
       err_all=chkfilen3(&dt);
    } else if(mode==1){
       err_all=chkfilen2(&dt);
    } else {
       err_all=chkfilen(&dt);
       if(err_all == -1) goto bypass; // no data file
       dt.err_all=err_all;
       /* summary out */
       if(imode>=1){
          summary_out(imode, logname1, &dt);
       }
       /* errored data file keep */
       if(keepmode>=1){
          if(err_all > 0){
              /* create file name to keep (original+".NNNN.err") */
              get_counter(f_counter, &number);
              strcpy(dt.keepfname,dt.fname);
              sprintf(cbuf,".%04d.err\0",number);
              strcat(dt.keepfname,cbuf);
              if(keepmode==1){   // rename
                  rename(dt.fname,dt.keepfname);
                  dt.keepstatus=1;
                  printf("data file renamed to %s\n",dt.keepfname);
              } else {           // copy
                  printf("data file copying to %s  .....\n",dt.keepfname);
#ifdef WIN32
                  sprintf(cbuf,"COPY %s %s\0",dt.fname, dt.keepfname);
#else
                  sprintf(cbuf,"cp %s %s\0",dt.fname, dt.keepfname);
#endif
                  i=system(cbuf); // 2011-02-08
                  dt.keepstatus=2;
                  printf("data file copied to %s\n",dt.keepfname);
              }

          }
       }

       /* log file out out */
       if(lmode>=1){
          err_log_out(lmode, logname2, &dt);
       }
bypass:;
    }
    cfinish1sec = clock();
    telapsed = (double)(cfinish1sec - cstart1sec) / CLOCKS_PER_SEC;
    printf("Time elapsed for processing is %f sec\n",telapsed);
    return 0;
}
